from abc import ABC, abstractmethod
from datetime import datetime

from pytz import BaseTzInfo, UTC


class Trigger(ABC):
    @property
    @abstractmethod
    def description(self) -> str:
        raise NotImplementedError()

    @abstractmethod
    def check(self) -> bool:
        raise NotImplementedError()

    def __str__(self) -> str:
        return self.description


class MinuteTrigger(Trigger):
    def __init__(self, minutes: int, tz: BaseTzInfo = UTC):
        assert minutes > 0

        self.minutes: int = minutes
        self.tz: BaseTzInfo = tz

    @property
    def description(self) -> str:
        if self.minutes == 1:
            return "every minute"

        return "every {minutes} minutes".format(
            minutes=self.minutes,
        )

    def check(self) -> bool:
        now = datetime.now(self.tz)

        return now.minute % self.minutes == 0

    def __str__(self) -> str:
        return self.description


class HourTrigger(Trigger):
    def __init__(self, hours: int, tz: BaseTzInfo = UTC):
        assert hours > 0

        self.hours: int = hours
        self.tz: BaseTzInfo = tz

    @property
    def description(self) -> str:
        tz = self.tz.zone

        if self.hours == 1:
            return "every hour (tz={tz})".format(
                tz=tz,
            )

        return "every {hours} hours (tz={tz})".format(
            hours=self.hours,
            tz=tz,
        )

    def check(self) -> bool:
        now = datetime.now(self.tz)

        if now.minute != 0:
            return False

        return now.hour % self.hours == 0

    def __str__(self) -> str:
        return self.description


class DayTrigger(Trigger):
    def __init__(self, hour: int, minute: int, tz: BaseTzInfo = UTC):
        assert 0 <= hour < 24
        assert 0 <= minute < 60

        self.hour: int = hour
        self.minute: int = minute
        self.tz: BaseTzInfo = tz

    @property
    def description(self) -> str:
        return "every day @ {hour:02d}:{minute:02d} (tz={tz})".format(
            hour=self.hour,
            minute=self.minute,
            tz=self.tz.zone,
        )

    def check(self) -> bool:
        now = datetime.now(self.tz)

        return now.hour == self.hour and now.minute == self.minute



import logging
from abc import ABC, abstractmethod
from datetime import datetime
from typing import Dict
from uuid import uuid4, UUID

from taskshooter.config import DEBUG
from .trigger import Trigger

logger = logging.getLogger(__name__)


class Task(ABC):
    def __init__(self, name: str, trigger: Trigger, emoji: str = None, metadata: Dict[str, object] = None):
        self.id: UUID = None
        self.name: str = name
        self.trigger: Trigger = trigger
        self.emoji: str = emoji
        self.metadata: Dict[str, object] = metadata or {}

        self.stated_at: datetime = None
        self.finished_at: datetime = None

    @abstractmethod
    def execute(self, force: bool = False):
        raise NotImplementedError()

    def run(self, force: bool = False, manual: bool = False):
        if not (force or manual):
            if not self.trigger.check():
                return

        self.id = uuid4()

        self.info("running task...")

        self.stated_at = datetime.now()
        self.pre_run()

        try:
            self.execute(force)
            self.finished_at = datetime.now()
            self.info(f"completed successfully in {self.runtime}ms")

            self.post_run(manual=manual, exception=None)
        except Exception as exception:
            self.exception(exception)
            self.finished_at = datetime.now()

            self.post_run(manual=manual, exception=exception)

        self.id = None

    def pre_run(self):
        pass

    def post_run(self, manual: bool = False, exception: Exception = None):
        pass

    @property
    def is_running(self) -> bool:
        return bool(self.stated_at and not self.finished_at)

    @property
    def runtime(self) -> float:
        if not self.finished_at:
            return None

        return int((self.finished_at - self.stated_at).microseconds / 1000)

    # logging
    def log(self, level: int, message: str, exception: Exception = None):
        logger.log(level, f"[%s] %s > %s", str(self.id), str(self), message, exc_info=exception)

        if DEBUG:
            print(f"[{self.id}] {self} > {message}")

    def debug(self, message: str, exception: Exception = None):
        self.log(logging.DEBUG, message, exception)

    def info(self, message: str, exception: Exception = None):
        self.log(logging.INFO, message, exception)

    def warning(self, message: str, exception: Exception = None):
        self.log(logging.WARNING, message, exception)

    def error(self, message: str, exception: Exception = None):
        self.log(logging.ERROR, message, exception)

    def exception(self, exception: Exception):
        self.error(str(exception), exception)

    def __repr__(self):
        if self.emoji:
            return f"{self.emoji} {self.name}"
        else:
            return f"{self.name}"



import logging
from concurrent.futures import ThreadPoolExecutor
from time import time, sleep

from taskshooter.config import DEBUG
from .task import Task

logger = logging.getLogger(__name__)


class Scheduler:
    def __init__(self, tasks: list[Task] = None):
        self.tasks = tasks or []

    def run(self):
        self.show()

        while True:
            self.nap()
            self.workwork()

    def add(self, task: Task):
        self.tasks.append(task)

    def show(self):
        self.info("Scheduled tasks:")

        for task in self.tasks:
            self.info(f" * {task}: {task.trigger.description}")

    def workwork(self):
        with ThreadPoolExecutor() as executor:
            executor.map(lambda task: task.run(), self.tasks)

    def nap(self):
        seconds = 60 - time() % 60
        self.debug("💤 sleeping")
        sleep(seconds)

    # logging
    def log(self, level: int, message: str, exception: Exception = None):
        logger.log(level, message, exc_info=exception)

        if DEBUG:
            print(message)

    def debug(self, message: str, exception: Exception = None):
        self.log(logging.DEBUG, message, exception)

    def info(self, message: str, exception: Exception = None):
        self.log(logging.INFO, message, exception)

    def warning(self, message: str, exception: Exception = None):
        self.log(logging.WARNING, message, exception)

    def error(self, message: str, exception: Exception = None):
        self.log(logging.ERROR, message, exception)

    def exception(self, exception: Exception):
        self.error(str(exception), exception)



from pytz import UTC

from .trigger import MinuteTrigger, HourTrigger, DayTrigger

minutes_1 = MinuteTrigger(1)
minutes_5 = MinuteTrigger(5)
minutes_10 = MinuteTrigger(10)
minutes_15 = MinuteTrigger(15)
minutes_20 = MinuteTrigger(20)
minutes_30 = MinuteTrigger(30)
minutes_60 = MinuteTrigger(60)

hours_1 = HourTrigger(1)
hours_2 = HourTrigger(2)
hours_3 = HourTrigger(3)
hours_6 = HourTrigger(6)
hours_12 = HourTrigger(12)
hours_24 = HourTrigger(24)

utc_0000 = DayTrigger(hour=0, minute=0, tz=UTC)
utc_0600 = DayTrigger(hour=6, minute=0, tz=UTC)
utc_1200 = DayTrigger(hour=12, minute=0, tz=UTC)
utc_1800 = DayTrigger(hour=18, minute=0, tz=UTC)
utc_2359 = DayTrigger(hour=23, minute=59, tz=UTC)
